@@ -1,46 +1,33 @@
-"""Client."""
+"""Matter Client implementation."""
 from __future__ import annotations
 
 import asyncio
 import logging
-import pprint
 from types import TracebackType
 from typing import TYPE_CHECKING, Any, Callable, Dict, Final, Optional, cast
 import uuid
 
-from aiohttp import ClientSession, ClientWebSocketResponse, WSMsgType, client_exceptions
+from aiohttp import ClientSession
 
-from ..common.helpers.json import json_dumps, json_loads
-from ..common.helpers.util import (
-    chip_clusters_version,
-    dataclass_from_dict,
-    dataclass_to_dict,
-    parse_message,
-)
-from ..common.models.api_command import APICommand
-from ..common.models.events import EventType
-from ..common.models.message import (
+from matter_server.common.errors import ERROR_MAP
+
+from ..common.helpers.util import dataclass_from_dict, dataclass_to_dict
+from ..common.models import (
+    APICommand,
     CommandMessage,
     ErrorResultMessage,
     EventMessage,
+    EventType,
+    MatterNodeData,
     MessageType,
     ResultMessageBase,
+    ServerDiagnostics,
     ServerInfoMessage,
     SuccessResultMessage,
 )
-from ..common.models.node import MatterAttribute, MatterNode
-from ..common.models.server_information import ServerInfo
-from .const import MIN_SCHEMA_VERSION
-from .exceptions import (
-    CannotConnect,
-    ConnectionClosed,
-    ConnectionFailed,
-    FailedCommand,
-    InvalidMessage,
-    InvalidServerVersion,
-    InvalidState,
-    NotConnected,
-)
+from .connection import MatterClientConnection
+from .exceptions import ConnectionClosed, InvalidServerVersion, InvalidState
+from .models.node import MatterNode
 
 if TYPE_CHECKING:
     from chip.clusters.Objects import ClusterCommand
@@ -53,20 +40,22 @@ class MatterClient:
 
     def __init__(self, ws_server_url: str, aiohttp_session: ClientSession):
         """Initialize the Client class."""
-        self.ws_server_url = ws_server_url
+        self.connection = MatterClientConnection(ws_server_url, aiohttp_session)
         self.logger = logging.getLogger(__package__)
-        self.aiohttp_session = aiohttp_session
-        # server info is retrieved on connect
-        self.server_info: ServerInfo | None = None
-        self._ws_client: ClientWebSocketResponse | None = None
-        self._loop = asyncio.get_running_loop()
         self._nodes: Dict[int, MatterNode] = {}
         self._result_futures: Dict[str, asyncio.Future] = {}
-        self._subscribers: dict[str, list[Callable]] = {}
+        self._subscribers: dict[str, list[Callable[[EventType, Any], None]]] = {}
+        self._stop_called: bool = False
+        self._loop: asyncio.AbstractEventLoop | None = None
+
+    @property
+    def server_info(self) -> ServerInfoMessage | None:
+        """Return info of the server we're currently connected to."""
+        return self.connection.server_info
 
     def subscribe(
         self,
-        callback: Callable[[EventType, Optional[Any]], None],
+        callback: Callable[[EventType, Any], None],
         event_filter: Optional[EventType] = None,
         node_filter: Optional[int] = None,
         attr_path_filter: Optional[str] = None,
@@ -104,46 +93,33 @@ class MatterClient:
 
         return unsubscribe
 
-    @property
-    def connected(self) -> bool:
-        """Return if we're currently connected."""
-        return self._ws_client is not None and not self._ws_client.closed
-
     async def get_nodes(self) -> list[MatterNode]:
         """Return all Matter nodes."""
-        if self._nodes:
-            # if start_listening is called this dict will be kept up to date
-            return list(self._nodes.values())
-        data = await self.send_command(APICommand.GET_NODES)
-        return [dataclass_from_dict(MatterNode, x) for x in data]
+        return list(self._nodes.values())
 
     async def get_node(self, node_id: int) -> MatterNode:
         """Return Matter node by id."""
-        if node_id in self._nodes:
-            # if start_listening is called this dict will be kept up to date
-            return self._nodes[node_id]
-        data = await self.send_command(APICommand.GET_NODE, node_id=node_id)
-        return dataclass_from_dict(MatterNode, data)
+        return self._nodes[node_id]
 
-    async def commission_with_code(self, code: str) -> MatterNode:
+    async def commission_with_code(self, code: str) -> MatterNodeData:
         """
         Commission a device using QRCode or ManualPairingCode.
 
-        Returns full NodeInfo once complete.
+        Returns basic MatterNodeData once complete.
         """
         data = await self.send_command(APICommand.COMMISSION_WITH_CODE, code=code)
-        return dataclass_from_dict(MatterNode, data)
+        return dataclass_from_dict(MatterNodeData, data)
 
-    async def commission_on_network(self, setup_pin_code: int) -> MatterNode:
+    async def commission_on_network(self, setup_pin_code: int) -> MatterNodeData:
         """
         Commission a device already connected to the network.
 
-        Returns full NodeInfo once complete.
+        Returns basic MatterNodeData once complete.
         """
         data = await self.send_command(
             APICommand.COMMISSION_ON_NETWORK, setup_pin_code=setup_pin_code
         )
-        return dataclass_from_dict(MatterNode, data)
+        return dataclass_from_dict(MatterNodeData, data)
 
     async def set_wifi_credentials(self, ssid: str, credentials: str) -> None:
         """Set WiFi credentials for commissioning to a (new) device."""
@@ -181,15 +157,30 @@ class MatterClient:
         )
 
     async def send_device_command(
-        self, node_id: int, endpoint: int, command: ClusterCommand
+        self,
+        node_id: int,
+        endpoint_id: int,
+        command: ClusterCommand,
+        response_type: Any | None = None,
+        timed_request_timeout_ms: int | None = None,
+        interaction_timeout_ms: int | None = None,
     ) -> Any:
         """Send a command to a Matter node/device."""
-        payload = dataclass_to_dict(command)
+        try:
+            command_name = command.__class__.__name__
+        except AttributeError:
+            # handle case where only the class was provided instead of an instance of it.
+            command_name = command.__name__
         return await self.send_command(
             APICommand.DEVICE_COMMAND,
             node_id=node_id,
-            endpoint=endpoint,
-            payload=payload,
+            endpoint_id=endpoint_id,
+            cluster_id=command.cluster_id,
+            command_name=command_name,
+            payload=dataclass_to_dict(command),
+            response_type=response_type,
+            timed_request_timeout_ms=timed_request_timeout_ms,
+            interaction_timeout_ms=interaction_timeout_ms,
         )
 
     async def remove_node(self, node_id: int) -> None:
@@ -203,11 +194,12 @@ class MatterClient:
         **kwargs: Any,
     ) -> Any:
         """Send a command and get a response."""
-        if not self.server_info:
+        if not self.connection.connected or not self._loop:
             raise InvalidState("Not connected")
 
         if (
             require_schema is not None
+            and self.server_info is not None
             and require_schema > self.server_info.schema_version
         ):
             raise InvalidServerVersion(
@@ -222,7 +214,7 @@ class MatterClient:
         )
         future: asyncio.Future[Any] = self._loop.create_future()
         self._result_futures[message.message_id] = future
-        await self._send_message(message)
+        await self.connection.send_message(message)
         try:
             return await future
         finally:
@@ -251,93 +243,49 @@ class MatterClient:
             command=command,
             args=kwargs,
         )
-        await self._send_message(message)
+        await self.connection.send_message(message)
+
+    async def get_diagnostics(self) -> ServerDiagnostics:
+        """Return a full dump of the server (for diagnostics)."""
+        data = await self.send_command(APICommand.SERVER_DIAGNOSTICS)
+        return dataclass_from_dict(ServerDiagnostics, data)
 
     async def connect(self) -> None:
-        """Connect to the websocket server."""
-        if self._ws_client is not None:
-            raise InvalidState("Already connected")
-
-        self.logger.debug("Trying to connect")
-        try:
-            self._ws_client = await self.aiohttp_session.ws_connect(
-                self.ws_server_url,
-                heartbeat=55,
-                compress=15,
-                max_msg_size=0,
-            )
-        except (
-            client_exceptions.WSServerHandshakeError,
-            client_exceptions.ClientError,
-        ) as err:
-            raise CannotConnect(err) from err
-
-        # at connect, the server sends a single message with the server info
-        info = cast(ServerInfoMessage, await self._receive_message_or_raise())
-        self.server_info = info
-
-        # sdk version must match exactly
-        if info.sdk_version != chip_clusters_version():
-            await self._ws_client.close()
-            raise InvalidServerVersion(
-                f"Matter Server SDK version is incompatible: {info.sdk_version} "
-                f"version on the client is {chip_clusters_version()} "
-            )
-
-        # basic check for server schema version compatibility
-        if info.schema_version < MIN_SCHEMA_VERSION:
-            # server version is too low, raise exception
-            await self._ws_client.close()
-            raise InvalidServerVersion(
-                f"Matter Server schema version is incompatible: {info.schema_version} "
-                f"a version is required that supports at least api schema {MIN_SCHEMA_VERSION} "
-                " - update the Matter Server to a more recent version."
-            )
-        if self.server_info.schema_version > MIN_SCHEMA_VERSION:
-            # server version is higher than expected, log only
-            self.logger.warning(
-                "Matter Server detected with schema version %s "
-                "which is higher than the preferred api schema %s of this client"
-                " - you may run into compatibility issues.",
-                info.schema_version,
-                MIN_SCHEMA_VERSION,
-            )
-
-        self.logger.info(
-            "Connected to Matter Fabric %s (%s), Schema version %s, CHIP SDK Version %s",
-            info.fabric_id,
-            info.compressed_fabric_id,
-            info.schema_version,
-            info.sdk_version,
-        )
+        """Connect to the Matter Server (over Websockets)."""
+        self._loop = asyncio.get_running_loop()
+        if self.connection.connected:
+            # already connected
+            return
+        # NOTE: connect will raise when connecting failed
+        await self.connection.connect()
 
     async def start_listening(self, init_ready: asyncio.Event | None = None) -> None:
         """Start listening to the websocket (and receive initial state)."""
-        if not self._ws_client or not self.connected:
-            raise InvalidState("Not connected when start listening")
+        await self.connect()
 
         try:
             message = CommandMessage(
                 message_id=uuid.uuid4().hex, command=APICommand.START_LISTENING
             )
-            await self._send_message(message)
+            await self.connection.send_message(message)
             nodes_msg = cast(
-                SuccessResultMessage, await self._receive_message_or_raise()
+                SuccessResultMessage, await self.connection.receive_message_or_raise()
             )
             # a full dump of all nodes will be the result of the start_listening command
-            nodes = {
-                x["node_id"]: dataclass_from_dict(MatterNode, x)
+            # create MatterNode objects from the basic MatterNodeData objects
+            nodes = [
+                MatterNode(dataclass_from_dict(MatterNodeData, x))
                 for x in nodes_msg.result
-            }
-            self._nodes = nodes
+            ]
+            self._nodes = {node.node_id: node for node in nodes}
             # once we've hit this point we're all set
             self.logger.info("Matter client initialized.")
             if init_ready is not None:
                 init_ready.set()
 
             # keep reading incoming messages
-            while not self._ws_client.closed:
-                msg = await self._receive_message_or_raise()
+            while not self._stop_called:
+                msg = await self.connection.receive_message_or_raise()
                 self._handle_incoming_message(msg)
         except ConnectionClosed:
             pass
@@ -345,40 +293,12 @@ class MatterClient:
             await self.disconnect()
 
     async def disconnect(self) -> None:
-        """Disconnect the client."""
-        self.logger.debug("Closing client connection")
+        """Disconnect the client and cleanup."""
+        self._stop_called = True
         # cancel all command-tasks awaiting a result
         for future in self._result_futures.values():
             future.cancel()
-        # close the client if is still connected
-        if self._ws_client is not None and not self._ws_client.closed:
-            await self._ws_client.close()
-
-    async def _receive_message_or_raise(self) -> MessageType:
-        """Receive (raw) message or raise."""
-        assert self._ws_client
-        ws_msg = await self._ws_client.receive()
-
-        if ws_msg.type in (WSMsgType.CLOSE, WSMsgType.CLOSED, WSMsgType.CLOSING):
-            raise ConnectionClosed("Connection was closed.")
-
-        if ws_msg.type == WSMsgType.ERROR:
-            raise ConnectionFailed()
-
-        if ws_msg.type != WSMsgType.TEXT:
-            raise InvalidMessage(f"Received non-Text message: {ws_msg.type}")
-
-        try:
-            msg = parse_message(json_loads(ws_msg.data))
-        except TypeError as err:
-            raise InvalidMessage(f"Received unsupported JSON: {err}") from err
-        except ValueError as err:
-            raise InvalidMessage("Received invalid JSON.") from err
-
-        if self.logger.isEnabledFor(logging.DEBUG):
-            self.logger.debug("Received message:\n%s\n", pprint.pformat(ws_msg))
-
-        return msg
+        await self.connection.disconnect()
 
     def _handle_incoming_message(self, msg: MessageType) -> None:
         """
@@ -388,7 +308,6 @@ class MatterClient:
         """
         # handle result message
         if isinstance(msg, ResultMessageBase):
-
             future = self._result_futures.get(msg.message_id)
 
             if future is None:
@@ -399,9 +318,8 @@ class MatterClient:
                 future.set_result(msg.result)
                 return
             if isinstance(msg, ErrorResultMessage):
-                future.set_exception(
-                    FailedCommand(msg.message_id, str(msg.error_code.value))
-                )
+                exc = ERROR_MAP[msg.error_code]
+                future.set_exception(exc(msg.details))
                 return
 
         # handle EventMessage
@@ -419,16 +337,33 @@ class MatterClient:
 
     def _handle_event_message(self, msg: EventMessage) -> None:
         """Handle incoming event from the server."""
-        if msg.event == EventType.NODE_ADDED:
-            node = dataclass_from_dict(MatterNode, msg.data)
-            self._nodes[node.node_id] = node
-            self._signal_event(EventType.NODE_ADDED, node)
+        if msg.event in (EventType.NODE_ADDED, EventType.NODE_UPDATED):
+            # an update event can potentially arrive for a not yet known node
+            node_data = dataclass_from_dict(MatterNodeData, msg.data)
+            node = self._nodes.get(node_data.node_id)
+            if node is None:
+                event = EventType.NODE_ADDED
+                node = MatterNode(node_data)
+                self._nodes[node.node_id] = node
+            else:
+                event = EventType.NODE_UPDATED
+                node.update(node_data)
+            self._signal_event(event, data=node, node_id=node.node_id)
+            return
+        if msg.event == EventType.NODE_DELETED:
+            node_id = msg.data
+            self._nodes.pop(node_id, None)
+            self._signal_event(EventType.NODE_DELETED, data=node_id, node_id=node_id)
             return
         if msg.event == EventType.ATTRIBUTE_UPDATED:
-            attr = dataclass_from_dict(MatterAttribute, msg.data)
-            self._nodes[attr.node_id].attributes[attr.path] = attr
+            # data is tuple[node_id, attribute_path, new_value]
+            node_id, attribute_path, new_value = msg.data
+            self._nodes[node_id].update_attribute(attribute_path, new_value)
             self._signal_event(
-                EventType.ATTRIBUTE_UPDATED, attr, attr.node_id, attr.path
+                EventType.ATTRIBUTE_UPDATED,
+                data=new_value,
+                node_id=node_id,
+                attribute_path=attribute_path,
             )
             return
         # TODO: handle any other events ?
@@ -455,23 +390,6 @@ class MatterClient:
                     for callback in self._subscribers.get(key, []):
                         callback(event, data)
 
-    async def _send_message(self, message: CommandMessage) -> None:
-        """
-        Send a CommandMessage to the server.
-
-        Raises NotConnected if client not connected.
-        """
-        if not self.connected:
-            raise NotConnected
-
-        if self.logger.isEnabledFor(logging.DEBUG):
-            self.logger.debug("Publishing message:\n%s\n", pprint.pformat(message))
-
-        assert self._ws_client
-        assert isinstance(message, CommandMessage)
-
-        await self._ws_client.send_json(message, dumps=json_dumps)
-
     async def __aenter__(self) -> "MatterClient":
         """Initialize and connect the Matter Websocket client."""
         await self.connect()
@@ -485,5 +403,6 @@ class MatterClient:
 
     def __repr__(self) -> str:
         """Return the representation."""
-        prefix = "" if self.connected else "not "
-        return f"{type(self).__name__}(ws_server_url={self.ws_server_url!r}, {prefix}connected)"
+        url = self.connection.ws_server_url
+        prefix = "" if self.connection.connected else "not "
+        return f"{type(self).__name__}(ws_server_url={url!r}, {prefix}connected)"
