@@ -7,27 +7,25 @@ from collections import deque
 from datetime import datetime
 from functools import partial
 import logging
-from typing import TYPE_CHECKING, Any, Callable, Deque, Optional, Type, TypeVar, cast
+from typing import TYPE_CHECKING, Any, Callable, Deque, Type, TypeVar, cast
 
 from chip.ChipDeviceCtrl import CommissionableNode
-from chip.clusters import Attribute, ClusterCommand
+from chip.clusters import Attribute
+from chip.clusters.ClusterObjects import ALL_CLUSTERS, Cluster
 from chip.exceptions import ChipStackError
 
-from ..common.helpers.api import api_command
-from ..common.helpers.util import dataclass_from_dict
-from ..common.models.api_command import APICommand
-from ..common.models.error import (
+from ..common.const import SCHEMA_VERSION
+from ..common.errors import (
     NodeCommissionFailed,
     NodeInterviewFailed,
     NodeNotExists,
     NodeNotResolving,
 )
-from ..common.models.events import EventType
-from ..common.models.node import MatterAttribute, MatterNode
-from .const import SCHEMA_VERSION
+from ..common.helpers.api import api_command
+from ..common.helpers.util import create_attribute_path, dataclass_from_dict
+from ..common.models import APICommand, EventType, MatterNodeData
 
 if TYPE_CHECKING:
-
     from .server import MatterServer
 
 _T = TypeVar("_T")
@@ -52,10 +50,11 @@ class MatterDeviceController:
         # we keep the last events in memory so we can include them in the diagnostics dump
         self.event_history: Deque[Attribute.EventReadResult] = deque(maxlen=25)
         self._subscriptions: dict[int, Attribute.SubscriptionTransaction] = {}
-        self._nodes: dict[int, MatterNode] = {}
+        self._nodes: dict[int, MatterNodeData | None] = {}
         self.wifi_credentials_set: bool = False
         self.thread_credentials_set: bool = False
         self.compressed_fabric_id: int | None = None
+        self._interview_task: asyncio.Task | None = None
 
     @property
     def fabric_id(self) -> int:
@@ -72,18 +71,23 @@ class MatterDeviceController:
     async def start(self) -> None:
         """Handle logic on controller start."""
         # load nodes from persistent storage
-        nodes_data = self.server.storage.get(DATA_KEY_NODES, {})
-        for node_id_str, node_dict in nodes_data.items():
+        nodes: dict[str, dict] = self.server.storage.get(DATA_KEY_NODES, {})
+        for node_id_str, node_dict in nodes.items():
             node_id = int(node_id_str)
-            node = dataclass_from_dict(MatterNode, node_dict)
+            # invalidate node data if schema mismatch,
+            # the node will automatically be scheduled for re-interview
+            if node_dict and node_dict.get("interview_version") != SCHEMA_VERSION:
+                node = None
+            else:
+                node = dataclass_from_dict(MatterNodeData, node_dict)
+                # always mark node as unavailable at startup until subscriptions are ready
+                node.available = False
             self._nodes[node_id] = node
-            # make sure to start node subscriptions
-            try:
-                await self.subscribe_node(node_id)
-            except NodeNotResolving:
-                LOGGER.warning("Node %s is not resolving, skipping...", node_id)
-        # create task to check for nodes that need any re(interviews)
-        self._schedule_interviews()
+        # setup subscriptions and (re)interviews as task in the background
+        # as we do not want it to block our startup
+        self._interview_task = asyncio.create_task(
+            self._check_subscriptions_and_interviews()
+        )
         LOGGER.debug("Loaded %s nodes", len(self._nodes))
 
     async def stop(self) -> None:
@@ -96,19 +100,23 @@ class MatterDeviceController:
         LOGGER.debug("Stopped.")
 
     @api_command(APICommand.GET_NODES)
-    def get_nodes(self) -> list[MatterNode]:
+    def get_nodes(self, only_available: bool = False) -> list[MatterNodeData]:
         """Return all Nodes known to the server."""
-        return [x for x in self._nodes.values() if x is not None]
+        return [
+            x
+            for x in self._nodes.values()
+            if x is not None and (x.available or not only_available)
+        ]
 
     @api_command(APICommand.GET_NODE)
-    def get_node(self, node_id: int) -> MatterNode:
+    def get_node(self, node_id: int) -> MatterNodeData:
         """Return info of a single Node."""
         node = self._nodes.get(node_id)
         assert node is not None, "Node does not exist or is not yet interviewed"
         return node
 
     @api_command(APICommand.COMMISSION_WITH_CODE)
-    async def commission_with_code(self, code: str) -> MatterNode:
+    async def commission_with_code(self, code: str) -> MatterNodeData:
         """
         Commission a device using QRCode or ManualPairingCode.
 
@@ -139,7 +147,7 @@ class MatterDeviceController:
         setup_pin_code: int,
         filter_type: int = 0,
         filter: Any = None,  # pylint: disable=redefined-builtin
-    ) -> MatterNode:
+    ) -> MatterNodeData:
         """
         Commission a device already connected to the network.
 
@@ -197,7 +205,7 @@ class MatterDeviceController:
         timeout: int = 300,
         iteration: int = 1000,
         option: int = 0,
-        discriminator: Optional[int] = None,
+        discriminator: int | None = None,
     ) -> int:
         """
         Open a commissioning window to commission a device present on this controller to another.
@@ -242,8 +250,9 @@ class MatterDeviceController:
         except ChipStackError as err:
             raise NodeInterviewFailed(f"Failed to interview node {node_id}") from err
 
+        is_new_node = node_id not in self._nodes
         existing_info = self._nodes.get(node_id)
-        node = MatterNode(
+        node = MatterNodeData(
             node_id=node_id,
             date_commissioned=existing_info.date_commissioned
             if existing_info
@@ -251,7 +260,7 @@ class MatterDeviceController:
             last_interview=datetime.utcnow(),
             interview_version=SCHEMA_VERSION,
             attributes=self._parse_attributes_from_read_result(
-                node_id, read_response.attributes
+                read_response.attributes
             ),
         )
 
@@ -263,19 +272,38 @@ class MatterDeviceController:
             value=node,
             force=not existing_info,
         )
-        if existing_info is None:
+        if is_new_node:
             # new node - first interview
             self.server.signal_event(EventType.NODE_ADDED, node)
+        else:
+            # existing node, signal node updated event
+            self.server.signal_event(EventType.NODE_UPDATED, node)
 
         LOGGER.debug("Interview of node %s completed", node_id)
 
     @api_command(APICommand.DEVICE_COMMAND)
     async def send_device_command(
-        self, node_id: int, endpoint: int, payload: ClusterCommand
+        self,
+        node_id: int,
+        endpoint_id: int,
+        cluster_id: int,
+        command_name: str,
+        payload: dict,
+        response_type: Any | None = None,
+        timed_request_timeout_ms: int | None = None,
+        interaction_timeout_ms: int | None = None,
     ) -> Any:
         """Send a command to a Matter node/device."""
+        cluster_cls: Cluster = ALL_CLUSTERS[cluster_id]
+        command_cls = getattr(cluster_cls.Commands, command_name)
+        command = dataclass_from_dict(command_cls, payload)
         return await self.chip_controller.SendCommand(
-            nodeid=node_id, endpoint=endpoint, payload=payload
+            nodeid=node_id,
+            endpoint=endpoint_id,
+            payload=command,
+            responseType=response_type,
+            timedRequestTimeoutMs=timed_request_timeout_ms,
+            interactionTimeoutMs=interaction_timeout_ms,
         )
 
     @api_command(APICommand.REMOVE_NODE)
@@ -304,12 +332,17 @@ class MatterDeviceController:
                 f"Node {node_id} does not exist or has not been interviewed."
             )
         assert node_id not in self._subscriptions, "Already subscribed to node"
-        LOGGER.debug("Setup subscription for node %s", node_id)
+        node_logger = LOGGER.getChild(str(node_id))
+        node_logger.debug("Setting up subscriptions...")
+
+        node = cast(MatterNodeData, self._nodes[node_id])
 
         try:
             await self._call_sdk(self.chip_controller.ResolveNode, nodeid=node_id)
         except ChipStackError as err:
+            node.available = False
             raise NodeNotResolving(f"Failed to resolve node {node_id}") from err
+
         # we follow the pattern of apple and google here and
         # just do a wildcard subscription for all clusters and properties
         # the client will handle filtering of the events.
@@ -325,10 +358,9 @@ class MatterDeviceController:
         ) -> None:
             assert self.server.loop is not None
             new_value = transaction.GetAttribute(path)
-            LOGGER.debug("attribute updated -- %s - new value: %s", path, new_value)
-            node = self._nodes[node_id]
-            attr = node.attributes[str(path.Path)]
-            attr.value = new_value
+            node_logger.debug("Attribute updated: %s - new value: %s", path, new_value)
+            attr_path = str(path.Path)
+            node.attributes[attr_path] = new_value
 
             # schedule save to persistent storage
             self.server.storage.set(
@@ -339,7 +371,10 @@ class MatterDeviceController:
 
             # This callback is running in the CHIP stack thread
             self.server.loop.call_soon_threadsafe(
-                self.server.signal_event, EventType.ATTRIBUTE_UPDATED, attr
+                self.server.signal_event,
+                EventType.ATTRIBUTE_UPDATED,
+                # send data as tuple[node_id, attribute_path, new_value]
+                (node_id, attr_path, new_value),
             )
 
         def event_callback(
@@ -348,7 +383,7 @@ class MatterDeviceController:
         ) -> None:
             # pylint: disable=unused-argument
             assert self.server.loop is not None
-            LOGGER.debug("received node event: %s", data)
+            node_logger.debug("Received node event: %s", data)
             self.event_history.append(data)
             # TODO: This callback does not seem to fire ever or my test devices do not have events
             self.server.loop.call_soon_threadsafe(
@@ -359,7 +394,7 @@ class MatterDeviceController:
             chipError: int, transaction: Attribute.SubscriptionTransaction
         ) -> None:
             # pylint: disable=unused-argument, invalid-name
-            LOGGER.error("Got error from node: %s", chipError)
+            node_logger.error("Got error from node: %s", chipError)
 
         def resubscription_attempted(
             transaction: Attribute.SubscriptionTransaction,
@@ -367,19 +402,25 @@ class MatterDeviceController:
             nextResubscribeIntervalMsec: int,
         ) -> None:
             # pylint: disable=unused-argument, invalid-name
-            LOGGER.debug(
-                "Previous subscription failed with Error: %s - re-subscribing in %s ms...",
+            node_logger.debug(
+                "Previous subscription failed with Error: %s, re-subscribing in %s ms...",
                 terminationError,
                 nextResubscribeIntervalMsec,
             )
-            # TODO: update node status to unavailable
+            # mark node as unavailable and signal consumers
+            if node.available:
+                node.available = False
+                self.server.signal_event(EventType.NODE_UPDATED, node)
 
         def resubscription_succeeded(
             transaction: Attribute.SubscriptionTransaction,
         ) -> None:
             # pylint: disable=unused-argument, invalid-name
-            LOGGER.debug("Subscription succeeded")
-            # TODO: update node status to available
+            node_logger.debug("Re-Subscription succeeded")
+            # mark node as available and signal consumers
+            if not node.available:
+                node.available = True
+                self.server.signal_event(EventType.NODE_UPDATED, node)
 
         sub.SetAttributeUpdateCallback(attribute_updated_callback)
         sub.SetEventUpdateCallback(event_callback)
@@ -387,6 +428,11 @@ class MatterDeviceController:
         sub.SetResubscriptionAttemptedCallback(resubscription_attempted)
         sub.SetResubscriptionSucceededCallback(resubscription_succeeded)
         self._subscriptions[node_id] = sub
+        # if we reach this point, it means the node could be resolved
+        # and the initial subscription succeeded, mark the node available.
+        node.available = True
+        node_logger.debug("Subscription succeeded")
+        self.server.signal_event(EventType.NODE_UPDATED, node)
 
     def _get_next_node_id(self) -> int:
         """Return next node_id."""
@@ -407,47 +453,70 @@ class MatterDeviceController:
             ),
         )
 
-    async def _check_interviews(self) -> None:
-        """Check if any nodes need to be (re)interviewed."""
-        assert self.server.loop is not None
-        # Reinterview all nodes that have an outdated schema
-        # and/or have been interviewed more than 30 days ago.
-        for node in self._nodes.values():
+    async def _check_subscriptions_and_interviews(self) -> None:
+        """Run subscriptions (and interviews) for known nodes."""
+        for node_id, node in self._nodes.items():
+            # (re)interview node (only) if needed
             if (
-                SCHEMA_VERSION > node.interview_version
+                node is None
+                or node.interview_version < SCHEMA_VERSION
                 or (datetime.utcnow() - node.last_interview).days > 30
             ):
-                await self.interview_node(node.node_id)
-        # reschedule self to run every hour
-        self.server.loop.call_later(3600, self._schedule_interviews)
+                try:
+                    await self.interview_node(node_id)
+                except NodeInterviewFailed as err:
+                    LOGGER.warning(
+                        "Unable to interview Node %s, we will retry later in the background.",
+                        node_id,
+                        exc_info=err,
+                    )
+                    continue
 
-    def _schedule_interviews(self) -> None:
-        """Schedule interviews."""
-        asyncio.create_task(self._check_interviews())
+            # setup subscriptions for the node
+            if node_id in self._subscriptions:
+                continue
+            try:
+                await self.subscribe_node(node_id)
+            except NodeNotResolving as err:
+                # If the node is unreachable on the network now,
+                # it will throw a NodeNotResolving exception, catch this,
+                # log this and just try to resolve this node in the next run.
+                LOGGER.warning(
+                    "Unable to contact Node %s,"
+                    " we will retry later in the background.",
+                    node_id,
+                    exc_info=err,
+                )
+
+        # reschedule self to run every hour
+        def _schedule() -> None:
+            """Schedule task."""
+            self._interview_task = asyncio.create_task(
+                self._check_subscriptions_and_interviews()
+            )
+
+        loop = cast(asyncio.AbstractEventLoop, self.server.loop)
+        loop.call_later(3600, _schedule)
 
     @staticmethod
     def _parse_attributes_from_read_result(
-        node_id: int, attributes: dict[int, dict[Type, dict[Type, Any]]]
-    ) -> dict[str, MatterAttribute]:
+        read_result: dict[int, dict[Type, dict[Type, Any]]]
+    ) -> dict[str, Any]:
         """Parse attributes from ReadResult."""
         result = {}
-        for endpoint, cluster_dict in attributes.items():
-            # read result output is in format {endpoint: {ClusterClass: {AttributeClass: value}}}
-            # we parse this to our own much more useable format
+        for endpoint, cluster_dict in read_result.items():
+            # read result output is in format
+            # {endpoint: {ClusterClass: {AttributeClass: value}}}
             for cluster_cls, attr_dict in cluster_dict.items():
                 for attr_cls, attr_value in attr_dict.items():
                     if attr_cls == Attribute.DataVersion:
                         continue
-                    attr = MatterAttribute(
-                        node_id=node_id,
-                        endpoint=endpoint,
-                        cluster_id=cluster_cls.id,
-                        cluster_type=cluster_cls,
-                        cluster_name=cluster_cls.__name__,
-                        attribute_id=attr_cls.attribute_id,
-                        attribute_type=attr_cls,
-                        attribute_name=attr_cls.__name__,
-                        value=attr_value,
+                    # we are only interested in the raw values and let the client
+                    # match back from the id's to the correct cluster/attribute classes
+                    # attributes are stored in form of AttributePath:
+                    # ENDPOINT/CLUSTER_ID/ATTRIBUTE_ID
+                    attribute_path = create_attribute_path(
+                        endpoint, cluster_cls.id, attr_cls.attribute_id
                     )
-                    result[attr.path] = attr
+                    result[attribute_path] = attr_value
         return result
